package node;

import filter.*;
import utils.*;
import capsule.*;
import config.*; // for the user DB
import clientserver.DecodingResult;
import clientserver.Result;

import java.util.*;

public  class Registry {
    // Allocates and deallocates the resources of the node.
    protected ConfigObject userDB; // Contains the customers and their profiles.
    // Ev. extend for several customers per policy generator...
    protected UserDBManager dbManager; 

    protected int maxAgents; // The maximum number of agents that can run here. It is
    // enforced via a limited number of server ports (see ServerPorts). It should
    // be done elsewhere...
    
    // Dynamic stuff:

    protected ServerPorts serverPorts;
    // Contains the assigned port numbers for the T_comp servers. 
    //The runningAgents could also be used but this is less efficient.
    // The methods of the class are synchronized
 
    protected Hashtable runningAgents=new Hashtable(); // Contains the currently running agents and
    //their assigned ressources.
 
    protected int nrOfAgents=0;

    protected Node node;
    protected RessourceController controller; // Checks regularly the ressources.

    protected boolean congestion=false; // If too much cpu (ev. memory) is used by the agents.

    public Registry(Node n) {
	node=n;
	maxAgents=n.MAX_AGENTS();
	dbManager=new UserDBManager(n);
	serverPorts=new ServerPorts(n.T_SERVER_PORT(),maxAgents);
	if (node.ENABLE_RESSOURCE_CONTROL()) {

	    controller=new RessourceController(this, node.CONTROL_INTERVAL(), node.MAX_BURST_LEN(), node.INTOLERABLE_LOSS(), node.MIN_SIGNIFICANT_EXEC_LEN(), node.MAX_TIME_2_NORMAL());
	    (new Thread(controller)).start();
	}
    }


    public synchronized int freePlaces() {
	return maxAgents-nrOfAgents;
    }

    public synchronized AARecord authorisationFor(String senderID, DecodingResult dec) {
	// Authorises the an execution request by generating an individual policy.
	// Also does the authentication if required by the user profile. 

	Object o=userDB.get(senderID);
	if(o==null) {
	    return new AARecord("Authorisation failed because user "+senderID+" is unknown!");
	}
	Profile userprofile=(Profile) o;
	Policy policy=userprofile.policyGen().getPolicyFor(userprofile); // Instantiate the policy for
	// this particular profile.
	if (policy.mustAuthenticate()) {
	    if(!dec.isSigned())
		return new AARecord("Authentication missing but required for user "+senderID+"!");
	    if(!dec.signedBy().equals(senderID)) 
		return new AARecord("Signed by "+dec.signedBy()+"instead of "+senderID+"!");
	}
	//System.out.println("Found profile for "+senderID+"\n"+userprofile.toString());
	if(congestion) { // Is this at the right place here?
	    // Here, the priorities could be checked.
	    return new AARecord("Authorisation failed because the node is currently congested.");
	}

	return new AARecord(policy);
    }

    public synchronized boolean registerAgent(RessourceUsage ru) {
	if(nrOfAgents>=maxAgents){
	    System.err.println("Error in registry: attempt to register too many agents. Synchrro-Problem!");
	    serverPorts.free(ru.serverPort());
	    return false;
	} else {
	    if(sameAgentIsRunning(ru.executionEnvironment())) {
		serverPorts.free(ru.serverPort()); // The ExecRequHandler has already assigned a port.
		return false; // To avoid breakdown when the agent floods the net.
	    }
	    runningAgents.put(ru.executionEnvironment().agent(), ru);
	    nrOfAgents++;
	    // also inform the scheduler/online resource controler here...
	    return true;
	}
    }

    public boolean sameAgentIsRunning(ExecutionEnvironment ev) {
	RessourceUsage ru;
	for(Enumeration e=runningAgents.elements();e.hasMoreElements(); ) {
	    ru=(RessourceUsage) e.nextElement();
	    if (ru.executionEnvironment().sameAgent(ev)) return true;
	}
	return false;
    }

    synchronized void stopAgent(Agent agent, Result finalMessage) {
	// The official method for stopping the whole agent externally.
	// FinalMessage may also be null.
	Object o=runningAgents.get(agent);
	if (o== null) {
	    System.err.println("Error: registry was requested to delete unknown agent: "+agent+"\nReason: "+finalMessage);
	    return;
	}
	nrOfAgents--;
	RessourceUsage ru=(RessourceUsage) o;
	System.out.println("Deregistering now agent: "+agent);
	runningAgents.remove(agent);
	ru.executionEnvironment().dispose(finalMessage); // Stops the queue stuff, too.
	serverPorts.free(ru.serverPort()); // deallocate the port number.
	// kill the agentwrapper thread here, in case it is not dead.
	ru.updateTotal(); // So that all totals are ok (for feedback/display).
	System.out.println("Registry says: pure agent running time was: "+ru.totalExecutionTime());
    }
    

    public int getServerPortNumber() {
	return serverPorts.assignAfreeNr(); 
    }

    public Vector runningAgentsOf(String owner) {
	// for answering queries.
	Vector ret=new Vector();
	RessourceUsage ev;
	for(Enumeration e=runningAgents.elements();e.hasMoreElements(); ) {
	    ev=(RessourceUsage) e.nextElement();
	    if (owner.equals(ev.executionEnvironment().owner())) ret.addElement(ev);
	}
	return ret;
    }

    synchronized Enumeration getRunningAgents() {
	// For RessourceController
	return runningAgents.elements();
    }

    void setCongestion(boolean c) {
	congestion=c;
    }

    int nrOfAgents() {
	return nrOfAgents;
    }

    public void killAllAgents() {
	RessourceUsage ev;
	for(Enumeration e=runningAgents.elements();e.hasMoreElements(); ) {
	    ev=(RessourceUsage) e.nextElement();
	    stopAgent(ev.executionEnvironment().agent(), null);
	}
    }

    protected void configRegistry() {
	// Will read the userDB from a file. Either Serialize or a special
	// parser can be used. Which one, is defined in node.Node.
	userDB=dbManager.defaultConfig();
    }
}
