package config;

import java.io.*;
import java.util.*;

import filter.*;
import node.*;

public class ProfileParser {
    // Reads and writes user profiles from/to a user readable file.
    // Format: 'username : homeNets : ProfileGeneratorName : nodeServicenumbers'

    protected static final String separator=":";
    protected StringTokenizer parser; // For splitting between the separators.
    protected StringTokenizer classNameParser; // Used in getGeneratorName.
    protected BufferedReader input=null;
    protected BufferedWriter output=null;

    protected FilterEntryParser homeNetParser=new FilterEntryParser();
    // The homenet is parsed with this parser already used by the GUI.
    protected Hashtable generators=new Hashtable(); // All known generator
    // names. Add new generators here.
    protected Vector services=new Vector(); // All non-default node service nrs.

    public ProfileParser(Node n) {
	// Add all known policy generators here.
	// The format uses the class name to find the object.
	// Note: we could also use class loading here... that would be
	// sweet!
	SuperPolicyGen spg= new SuperPolicyGen(n);
	CustomerPolicyGen cpg= new CustomerPolicyGen(n);
	AnonymousUserPolicyGen aupg= new AnonymousUserPolicyGen(n);
	TestPolicyGen tpg= new TestPolicyGen(n);
	generators.put(getGeneratorName(spg),spg);
	generators.put(getGeneratorName(cpg),cpg);
	generators.put(getGeneratorName(tpg),tpg);
	generators.put(getGeneratorName(aupg),aupg);
    }

    public void setReader(Reader r) {
	// Should be called before 'read'.
	input=new BufferedReader(r);
    }

    public void setWriter(Writer w) {
	// Should be called before 'write'.
	output=new BufferedWriter(w);
    }

    public void write(Profile p) {
	try {
	    if(output==null) {
		System.err.println("ProfileParser error: no writer set.");
		return;
	    }
	    String s="\t"+separator+"\t";
	    output.write(p.identity());
	    output.write(s);
	    output.write(homeNetParser.printAddressRanges(p.homeNets()));
	    output.write(s);
	    output.write(getGeneratorName(p.policyGen()));
	    output.write(s);
	    output.write(serviceNrs(p.services()));
	    output.write('\n');
	    output.flush();
	} catch (IOException e) {
	    System.err.println("ProfileParser write error: "+e);
	}
    }

   

    public Profile read() {
	try {
	    if(input==null) {
		System.err.println("ProfileParser read error: no reader set.");
		return null;
	    }
	    String line=input.readLine();
	    if (line==null) return null; // End of data.
	    return parseLine(line);
	} catch (Exception e) {
	    System.err.println("ProfileParser error: "+e);
	    e.printStackTrace();
	}
	return null;
    }

    protected Profile parseLine(String line) {
	// Parses one line of input into a profile object.
	Profile ret=null;
	parser=new StringTokenizer(line,separator+"\n");
	String id=null;

	ListOfRanges homeNets=null;
	PolicyGenerator generator=null;

	id=giveNextToken(parser);

	homeNets=homeNetParser.parseAddressRanges(giveNextToken(parser));
	generator=lookupGeneratorClass(giveNextToken(parser));
	services=parseServiceNrs(giveNextToken(parser));

	if ((id==null) ||(homeNets==null) || (generator==null)|| (services==null) ) {
	    System.err.println("ProfileParser error: parsing error in line: "+line);
	    return null;
	}
	return new Profile(id, homeNets, generator, services);
    }

    protected String giveNextToken(StringTokenizer parser) {
	String ret=null;
	try {
	    ret=parser.nextToken();
	    ret=ret.trim(); // eliminate leading/trailing whitespaces.
	} catch(Exception e) {}
	return ret;
    }

    protected PolicyGenerator lookupGeneratorClass(String n) {
	String name=n.trim();
	return (PolicyGenerator) generators.get(name);
    }

    protected Vector parseServiceNrs(String nrs) {
	Vector ret=new Vector();
	if(nrs==null) return ret;
	StringTokenizer tokenizer=new StringTokenizer(nrs);
	Integer i;
	String token;
	while (tokenizer.hasMoreTokens()) {
	    token=tokenizer.nextToken();
	    try {
		i= new Integer(Integer.parseInt(token));
		ret.addElement(i); 
	    } catch(NumberFormatException e) { 
		System.err.println("Not a service number "+token+" --> ignored!");
	    }
	}
	return ret;
    }

    protected Boolean parseBoolean(String s) {
	return Boolean.valueOf(s); 
    }

    protected String getGeneratorName(PolicyGenerator pg) {
	// The name is used as key in the 'generators' hashtable.
	String objName=pg.getClass().toString(); // The name of the class object.
	classNameParser=new StringTokenizer(objName,". \t\n");
	String token=null;
	while(classNameParser.hasMoreTokens()) {
	    token=classNameParser.nextToken();
	}
	//return classNameParser.nextToken(); // Extract away the object number, 
	// returning the pure class name.
	return token;
    }

    protected String serviceNrs(Vector srvs) {
	StringBuffer buf=new StringBuffer();
	Integer i;
	for(Enumeration e=srvs.elements(); e.hasMoreElements();) {
	    i=(Integer) e.nextElement();
	    buf.append(" ");
	    buf.append(i);
	}
	return new String(buf);
    }



}
