package filter;

import java.io.*;

import node.*; // for the IPPacket def.
import utils.*;

public class Filter implements Serializable {
    protected static int maxPacketLen=8120; // Longer packets will
    // not be dumped correctly!
    protected static String snifferHub="";  // To avoid duplicates when sniffing// hub: 0:0:1d:b9:8f:e7
    // on a hub (packet-2-hub and hub-2-dest) we sniff 5only things going out of the hub.
    // If there is no hub (switch) put an empty string here.


    protected ListOfRanges src;
    protected ListOfRanges dst;
    protected ListOfRanges port; // source or dest. Should we differentiate?
    protected ListOfRanges proto;
    protected Length dataLen; // Short packets (e.g. only header seems also to be a problem for tcpdump. Filter len is apparently +14 bytes (ethernet), so 
	// only ip header is e.g. -s 34 instead of 20 (on linux).
    // New meaning: length of ip payload to be snapped.
    protected Length packets;

    protected PortMatcher portMatcher=new PortMatcher();

    protected final String tcpDumpArgs="";

    public Filter(ListOfRanges s, ListOfRanges d,ListOfRanges sp, ListOfRanges p, int l, int pl) {
	initWith(s,d,sp,p, l, pl);
    }

    public Filter(ListOfRanges s, ListOfRanges d,ListOfRanges sp, ListOfRanges p, Length l, Length pl) {
	initWith(s,d,sp,p, l.getNr(), pl.getNr());
    }

    public Filter(int s, int d, int sp, int p, int l, int pl) {
	// convenient constructor for a filter without wildcards.
	// We could also use Range instead of List:
	initWith(new ListOfRanges(new Range(s)),
		 new ListOfRanges(new Range(d)),
		 new ListOfRanges(new Range(sp)),
		 new ListOfRanges(new Range(p)),
		 l, pl);
    }
    
    public Length getIPLen() {
	// returns the number of ip bytes
	return new Length(dataLen.getNr()) ;
    }

    public Length packets() {
	// returns the number of packets to be filtered.
	return new Length(packets.getNr()) ;
    }

    public ListOfRanges source() {return src;}
    public ListOfRanges destination() {return dst;}
    public ListOfRanges protocol() {return proto;}


    protected void initWith(ListOfRanges s, ListOfRanges d, ListOfRanges sp, ListOfRanges p, int dl, int pl) {
       	src=s;
	dst=d; 
       	port=sp;
	proto=p;
	checkNsetDataLen(dl);
	packets=new Length(pl);
    }

    public boolean equals(Filter f) {
	// if the same packets are matched
	if (!src.equals(f.src)) return false;
	if (!dst.equals(f.dst)) return false;
	if (!port.equals(f.port)) return false;
	if (!proto.equals(f.proto)) return false;
	if (!dataLen.equals(f.dataLen)) return false;
	if (!packets.equals(f.packets)) return false;
	return true;

    }

    public boolean match(IPPacket p) {
	// Only a coarse checking. The true matching is done in tcpdump
	if (!src.match(p.source())) return false;
	if (!dst.match(p.destination())) return false;
	if (!proto.match(p.proto())) return false;
	// ports are not necessarily defined.
	return portMatch(p);
    }

    public boolean portMatch(IPPacket p) {
	portMatcher.setPacket(p);
	return portMatcher.matches(port);
    }
    

    public Filter and(Filter f) {
	ListOfRanges s, d,sp, dp, p;
	Length l, pl;
	s=src.and(f.src);
	d=dst.and(f.dst);
	sp=port.and(f.port);
	p=proto.and(f.proto);

	l=dataLen.niAnd(f.dataLen); // The minimum. Zero is no problem; we will still get the header.
	pl=packets.niAnd(f.packets);
	return new Filter (s,d,sp,p,l.getNr(),pl.getNr());
    }


    public boolean isEmpty() {
	if (src.isEmpty()) return true;
	if (dst.isEmpty()) return true;
	if (port.isEmpty()) return true;
	if (proto.isEmpty()) return true;
	if (packets.getNr()==0) return true; 
	return false;
    }

    public String toString() {
	FilterEntryParser parser=new FilterEntryParser();// can write
	// a ListOfRanges of addresses in human readable form.
	
	StringBuffer buf=new StringBuffer();
	//buf.append("\nFilter:");
	buf.append("\nSource: ");
	buf.append(parser.printAddressRanges(src));
	buf.append("\nDestination: ");
	buf.append(parser.printAddressRanges(dst));
	buf.append("\nPort: ");
	buf.append(port.toString());
	buf.append("\nProtocol: ");
	buf.append(proto.toString());
	buf.append("\nData length requested: ");
	dataLen.appendTo(buf);
	buf.append("\nNr of packets requested: ");
	packets.appendTo(buf);
	return new String(buf);
    }

    public String toTcpDumpArgs() {
	StringBuffer buf=new StringBuffer(tcpDumpArgs);
	String tmp;
	//buf.append('\''); 
	if(!packets.meansAny()) {
	    buf.append(" -c ");
	    buf.append(packets.getNr());
	}

	buf.append(" -s ");
	if(dataLen.meansAny()) {
	    buf.append(maxPacketLen);
	} else {
	    buf.append(dataLen.getNr()+34); // 14 ether and 20 ip
	}
       
	if(!snifferHub.equals("")) { // Append ether filter if hub is used.
	    buf.append(" ether src ");
	    buf.append(snifferHub);
	    buf.append(" and");
	}
	buf.append(" ip");
	tmp=proto.tcpDumpProtoSpec();
	if(!tmp.equals("")) { // if proto specified.
	    buf.append(" and ");
	    buf.append(tmp);
	}
	tmp=port.tcpDumpPortSpec();
	if(!tmp.equals("")) { // if proto specified.
	    buf.append(" and ");
	    buf.append(tmp);
	}
	tmp=src.tcpDumpAddrSpec("src");
	if(!tmp.equals("")) { // if proto specified.
	    buf.append(" and ");
	    buf.append(tmp); 
	}
	tmp=dst.tcpDumpAddrSpec("dst");
	if(!tmp.equals("")) { // if proto specified.
	    buf.append(" and ");
	    buf.append(tmp); 
	}
	//buf.append('\''); 
	return new String(buf);
    }


    protected void checkNsetDataLen(int l) {
	if (checkDataLen(l))  dataLen=new Length(l);
	else dataLen=new Length(-1); // default; meaning we want all data.
    }

    public boolean checkDataLen(int l) {
	return (l<65536); // any negative is ok and treated as no filter
    }


    public static void main(String[] args) {
	// Testing
// 	All a=new All();

// 	int ia[]={1,2,4};
// 	int ib[]={1,7,9};
// 	List l1=new List(ia);
// 	List l2=new List(ib);

// 	Range r1=new Range(0,255);
// 	Range r2=new Range(0,1222);
   

// 	Filter f1=new Filter(r1,l1,l2,4,-1);
// 	Filter f2=new Filter(r2,r2,l1,-1,10000);
// 	System.out.println(f1);
// 	System.out.println(f2);
// 	System.out.println();

// 	System.out.println(f1.and(f2));
// 	System.out.println(f1.isEmpty());
// 	System.out.println(f1);
// 	System.out.println(f1.toTcpDumpArgs());



// 	int dst1, dst2;
// 	int[] dsts=new int[2];
// 	dsts[0]=AddressCalculator.parseIPaddress("130.92.65.25");
// 	dsts[1]=AddressCalculator.parseIPaddress("130.92.65.24");


// 	Range proto=new Range(0,10);
// 	Range from=new Range(AddressCalculator.parseIPaddress("130.92.64.11"),(short) 24);
// 	List to=new List(dsts);
//  	Filter f1=new Filter(from,to,proto,-1,4);
//  	System.out.println(f1);
//  	System.out.println(f1.toTcpDumpArgs());
	
    }
}
