package filter;

import utils.*;
import java.io.*;

public class  Range implements Serializable {
    protected int upper=0; 
    protected int lower=0; 

    public Range() {
	initWith(1,-1); // An empty range.
	// This is solely needed for the subclass All.java
    }

    public Range(int single) {
	initWith(single,single);
    }


    public Range(int low, int up) {
	initWith(low, up);
    }

    public Range(int addr, short sigBits) {
	// The network\len notation.
	if((sigBits<1)||(sigBits>32)) {
	    System.err.println("Significant bits should be between 1-32!");
	    initWith(1,-1); // An empty range.
	} else {
	    int notSig=( 32-sigBits);
	    int e=1<<notSig;
	    e--; //e == upper-lower.
	    int low=addr>>notSig; // erease the not significant bits.
	    low=low<<notSig;
	    initWith(low, low+e);
	}   
    }

    protected void initWith(int low, int up) {
	lower=low;
	upper=up;
    }

    // --- Accessors

    public int low() {
	return lower;
    }

    public int up() {
	return upper;
    }

    public int nrOfMatches() {
	return upper-lower+1;
    }

    // --- Set testing operations

    public boolean isEmpty() {
	return lower>upper;
    }


    public boolean match(int r) {
	return (r>=lower) && (r<=upper);
    }


    public Range and(Range e) {
	int rl=e.lower;
	int ru=e.upper;
	int min, max;
	max= (lower>rl) ? lower : rl; // max of the lower
	min=(upper<ru) ? upper: ru; // min of the upper

	return new Range(max, min);
    }

    public boolean isMaximal() {
	return (lower==Integer.MIN_VALUE) && (upper==Integer.MAX_VALUE);
    }

    // --- Helper for ListOfRanges

    // beware of empty ranges!
    public boolean smaller(Range o) {
	return upper<o.lower;
    }
    public boolean touching(Range o) {
	return ((lower<=o.upper) && (lower>=o.lower)) ||
	    ((o.lower<=upper) && (o.lower>=lower));
    }
    


    // --- Output to String

    public String toString() {
	StringBuffer buf=new StringBuffer();
	//buf.append(super.toString());
	buf.append(lower);
	buf.append('-');
	buf.append(upper);
	return new String(buf);
    }

    public String toNetNotation() {
	// Prints the address\length notation if possible
	StringBuffer buf=new StringBuffer();
	int add=lower; 
	int trailingZeroes=countTrailingZeroes(add);
	long e=nrOfMatches();
	if ((isAllOne(e- 1)) && (e <= ((long) 1<<trailingZeroes))) { // range can be converted
	    buf.append(AddressCalculator.toString(lower));
	    buf.append('/');
	    buf.append(32-countTrailingZeroes(e));
	} else {
	    buf.append("This range cannot be converted to net/len notation.");
	}
	return new String(buf);
    }

    public String tcpDumpProtoSpec() {
	StringBuffer buf=new StringBuffer();
	buf.append("((ip[9]>= ");
	buf.append(lower);
	buf.append(") && (ip[9]<= ");
	buf.append(upper);
	buf.append("))");
	return new String(buf);
    }

    public String tcpDumpAddrSpec(String direction) {
	String low=AddressCalculator.toString(lower);
	StringBuffer buf=new StringBuffer(direction);

	if (nrOfMatches()==1) {
	    buf.append(" host ");
	    buf.append(low);
	} else {
	    String up=toNetNotation(); // This can also fail!
	    buf.append(" net ");
	    buf.append(up);
	}
	return new String(buf);
    }


    // --- Notation helpers

    protected boolean isAllOne(long i) {
	while (i!=0) {
	    if (i%2==0) return false;
	    i>>=1;
	}
	return true;
    }

    protected int countTrailingZeroes(long i) {
	int c=0;
	while (i%2==0) {
	    i>>=1;
	    c++;
	    if (c==32) break;
	}
	return c;
    }

    public static void main(String[] args) {
	// Testing
// 	int o=13;
// 	int e=1;

// 	Wildcard r=new Range(1,10);
// 	Wildcard r2=new Range(3,17);
// 	Wildcard r3=new Range(3,17);
// 	Wildcard r4=new Range(1000000000,(short)24);

// 	All a=new All();
// 	int list[]= {0, 1, 2, 4, 17};
// 	int list2[]= {103, 104};
// 	int list4[]= {1,2,3,4,5,6,7,8,9,10};
// 	List l=new List(list);
// 	List l2=new List(list2);
// 	List l3=new List(list2);
// 	List l4=new List(list4); // equals r

// 	System.out.println(r2.and(a));
// 	System.out.println(r2.and(l));
// 	System.out.println(r2.and(r));

// 	System.out.println("TCP dump tests");
// 	System.out.println(r.tcpDumpProtoSpec());
// 	System.out.println(r4.tcpDumpAddrSpec("src"));
// 	System.out.println(a.tcpDumpProtoSpec());
// 	System.out.println(l.tcpDumpAddrSpec("dst"));
	

// 	System.out.println("Equality checks");
// 	System.out.println(r.equals(r2)); // false
// 	System.out.println(r2.equals(r3)); // true
// 	System.out.println(r.equals(l)); // false
// 	System.out.println(r.equals(l4)); // true
// 	System.out.println(r.equals(a)); // false
// 	System.out.println(a.equals(a)); // true
// 	System.out.println(a.equals(r)); // false
// 	System.out.println(a.equals(l)); // false
// 	System.out.println(l.equals(l2)); // false 
// 	System.out.println(l2.equals(l3)); // true
// 	System.out.println(l.equals(r)); // false
// 	System.out.println(l4.equals(r)); // true 


// 	System.out.println(a.match(o));
// 	System.out.println(r.match(o));
// 	System.out.println(r.match(e));
// 	System.out.println(a.and(r));
// 	System.out.println(r.and(a));
// 	System.out.println();
// 	System.out.println(l.and(a));
// 	System.out.println(l.and(l));
// 	System.out.println(l.and(r));
// 	System.out.println();
// 	System.out.println(r.and(l));
// 	System.out.println((l2.and(l)).isEmpty());
// 	System.out.println(r2.and(r));



    }
}
