package node;

import clientserver.DummyResult;
import clientserver.Result;
import utils.ObjectQueue;
import node.services.Service;

import capsule.*;
import filter.*;
import java.io.*;
import java.util.*;


public class AgentWrapper extends ServiceDelivery {
    // Encapsulates the thread that is running the agent.
    // This is the only class that directly calls methods of the agent.
    // Instead of this, a superclass of all agents could probably do the job
    // more easily (especially starting & stopping the clock), but the
    // classloader makes troubles.
    // Each call to the agent is sandwiched with clock calls to register
    // the running time. Security exceptions are caught and the agent
    // is subsequently terminated.

    protected  int criticalQlen;

    // Relays the packets to the agent.
    protected int nrOfPackets;
    protected ServiceChannel reply;

    protected Vector initVals; // For the initialisation of the agent
    protected filter.Filter filter;
  
    public AgentWrapper(Agent a, Vector initVals, ObjectQueue u, filter.Filter f , ServiceChannel reply, Node node) {
	init(node,a,u);
	this.initVals=initVals;
	nrOfPackets=f.packets().getNr();
	this.reply=reply;
	criticalQlen=node.CRITICAL_Q_LEN();
	filter=f;
    }


    public void run() {
	// Implements the life of the agent. 
	IPPacket ip;
	try {
	    startClock();
	    agent.putServiceChannel(reply);// The agent can now report to where it came from.
	    if(agent.init(initVals)) { // If init works, deliver the packets.
		stopClock();
		
		for (int i=0; (nrOfPackets<0) || (i<nrOfPackets); i++) {
		    // Note that a negative i means go on forever.
		    //if(stopping) break; // External stopping, even if All packets are requested. Now done with thread.stop()
		    ip=(IPPacket) queue.get(); // Blocks if no packet there.
		    //if(filter.match(ip))  {// Full refiltering. (vr filter bug).
			//if(filter.portMatch(ip)) // Refiltering for VR
			startClock();
			if(queue.used()>=criticalQlen) {
			    agent.emergencyHandle(ip); // The call to the agent.
			    usage.incEmergencyHandle();
			} else {
			    agent.handlePacket(ip); // The call to the agent.
			    usage.incHandledPackets();
			}
			stopClock();
			//}
		}
		    
		startClock();
		agent.stop(); // Could be done by the registry, but we want to measure the time...
		stopClock();
	    } else { // Initialisation failed
		//reply.request(new Result("Initialisation with:\n"+initVals+"\nFAILED.",true));
		System.err.println("Initialisation with:\n"+initVals+"\nFAILED.");
	    }

	} catch (SecurityException e) { // The agents actions may cause a security exception or class loader exceptions.
	    reactUponSecurityViolation(e);
	} catch (NoClassDefFoundError e) { // Usually this error should not be caught.
	    // However, the classloader can cause a security exception which will
	    // lead to a NoClassDefFoundError. So we have to translate that here.
	    System.err.println("-->No CLASS def found (in AgentWrapper): "+e);
	    reactUponSecurityViolation(new Exception("Class loader security error."));
	}
	
	signalStopToServiceChannel();
	//if(informRegistry) node.registry().stopAgent(agent, null); // Stopps the agent.
	System.out.println("Agent thread terminates self-guided.");

    }

    protected void  signalStopToServiceChannel() {
	ObjectQueue queue=reply.getQueue();
	if(!queue.full())  // Else this can block which can lead to a deadlock. 
	    queue.put(new Service(0,"Normalstopp"));
    }

}
