package node;

import java.io.*;
import java.util.*;

import node.services.*;
import capsule.*;
import clientserver.*;
import filter.*;
import utils.*;

public class ExecutionRequestHandler {
    // The server side

    protected ServerPOSenderReceiver posr;
    protected ErrorMessage error=null;
    protected Registry registry; // registers agent and their ressources, does the authorisation.
    protected Node node;

    public  ExecutionRequestHandler(Node n, String senderID, DecodingResult dec , ExecutionRequest req, ServerPOSenderReceiver p) {
	// The constructor does the whole job of setting up the execution environment and
	// registering the agent. 
	node=n;
	posr=p;
	registry=node.registry();

	handleRequest(senderID, dec,  req);
    }

    protected synchronized void handleRequest(String senderID, DecodingResult dec, ExecutionRequest req) {
	// The setup and start of the agent. The method must be synchronized.
	// This method incorporates most of the agent installation procedure.
	// -> factor out pieces, if time available.

	//System.out.println("Handling now: "+req);

	// authorisation: lookup profile, get resources (filter)
	Policy policy=authorise(senderID, dec);
	if(error != null) {
	    sendNack(error.toString());
	    return;
	}
	
	Agent agent=instantiateAgent(senderID, req); // prepare agent
	if(error != null) {
	    sendNack(error.toString());
	    return;
	}

	// prepare agent resources and exec env
	Filter f=policy.getFirstMatchingFilter(req.filter); // creation of the policed filter.
	if(f==null) {
	    sendNack("Requested filter is not acceptable.");
	    return;
	}

	//System.out.println("Using filter: "+f);

	int serverport=registry.getServerPortNumber(); // get a port to connect from the T-component.
	if(serverport<0) { 
	    sendNack("Not enough agent places free.");
	    return;
	}
	// Make sure this gets deregistered again.


	// Prepare the execution environment of the agent: 
	ObjectQueue in=new ObjectQueue(node.Q_LEN());
	ServiceChannel reply=new ServiceChannelV1(agent,  node);

	AgentWrapper wrapper=new AgentWrapper(agent, req.initValues, in, f, reply, node); 
	NodeServiceDelivery deliver=new NodeServiceDelivery(agent, node);
	Receiver receiver=new Receiver(serverport, in, f ,node,agent);
	//DummyReceiver  receiver=new DummyReceiver(serverport, in, f ,node,agent);
	//T_Configurator t_comp_client=null;
	T_Configurator t_comp_client=new  T_Configurator(node, serverport, f, receiver); //also inits the receiver.

	int hopCount=req.forwarding.getHopCount();
	ExecutionEnvironment ev= new ExecutionEnvironment(senderID,req.ID, agent, hopCount,  f,  reply, wrapper,deliver, receiver, posr, policy);
	RessourceUsage usage=new RessourceUsage(ev, policy, serverport);
	wrapper.setRessourceUsage(usage); // The wrapper updates the usage of ressources.
	deliver.setRessourceUsage(usage); // ..so does the service result delivery
	reply.setRessourceUsage(usage); //  and the service channel.


	if(!registry.registerAgent(usage)) {// Note, that the registry has full access 
	    sendNack("The same agent is already running!");// through the execution environment.
	    return;
	}
    
	// Start agent & resources (execution environment)

	reply.initServiceList(policy.services(),ev); // Installs node service handlers.
	ev.start(t_comp_client); // Starts all the threads etc.

	usage.setStartingTime(System.currentTimeMillis());

	(new Thread(t_comp_client)).start(); // start it after the wrapper (server)!
 
	posr.startForwarding(); // If all went well here, forwarding the agent(if requested).
	sendAck();
    }

    protected Policy authorise(String senderID, DecodingResult dec) {
	// Authentication and lookup the policy by asking the registry.

	AARecord aa=registry.authorisationFor(senderID, dec);

	if(aa.failed) {
	    error=new ErrorMessage(aa.errorMessage);
	    return null;
	} 
	return aa.policy;
    }


    protected Agent instantiateAgent(String id, ExecutionRequest req) {
	// Uses an agent class loader to load all classes necessary for the agent.
  	Agent o=null;
	
	try {
	    AgentClassLoaderDynamic acl=new AgentClassLoaderDynamic(id); // One loader per request (namespace protection).
	    error=acl.pushClasses(req.classNames, req.classes);
	    if (error != null) return null;
	    
	    Class AgentClass=acl.loadClass((String) (req.classNames.firstElement()), true);
	    // There should be a check here..

	    o= (Agent) AgentClass.newInstance();

       } catch(Exception e) {
	   System.err.println("Error in ExecutionRequestHandler.instantiateAgent(): ");
	   System.err.println(e+" "+ e.getMessage());
	   e.printStackTrace();
	   error=new ErrorMessage(e.toString());
	   return null;
       }
       return o;
    }


    protected void sendAck() {
	sendMessage(new Acknowledgement());
    }
    protected void sendNack(String reason) {
	sendMessage(new Nacknowledgement(reason));
    }

    protected void sendMessage(Message m) {
	posr.send(new  ProtocolObject(node.NODE_ID(), ProtoConsts.PLAIN ,  ProtoConsts.EXEC_ACK,  m));
	// Note that posr may change the encoding, to the same used in the request.
    }
}
