package node;

import clientserver.*;
import node.services.ServiceDatabase;
import config.*;
import topology.*;
import utils.PrimitiveTypeParser;

import java.net.*; // For dns lookup
import java.io.Serializable; // For saving as config stuff
import java.util.Vector; 

public class Node implements Serializable {
    // Starts a server that welcomes and executes agents. This class contains
    // the static values of the node configuration and some databases
    // (services, topology and registered agents).

    protected static final String defaultConfigName="./config/default"; // Default file for the configs that are accessed via
    // the node from all over the node package.
    
    protected Registry registry; // Contains the dynamic information of the node, 
    // e.g. what agents are running and their ressource consumption.
    protected NeighborLookupTable neighborLookup; // Has access to a central conf file.
    protected NodeLookupTable nodeLookup; // Has access to a central conf file.
    protected RoutingTable routingTable; // To identify the next node towards an IP.
    protected NodeRecord aboutMe; // The location/topo info of this node.
    
    protected NodeConfigManager configManager; // The loader of the config file.
    protected ConfigObject configurations ; // The runtime store of the configs.
    protected PrimitiveTypeParser ptParser=new PrimitiveTypeParser();
    // Configs are stored as Objects. ptParser helps to convert them to e.g.ints.

    protected ServiceDatabase nodeServices;
    protected Timer timer=new Timer(); // For performance tests.

    public Node(String confFile) {
	configManager=new NodeConfigManager(confFile);
	configurations=configManager.defaultConfig(); // Load the config parameters.
	initNodeInfo(); // The node info is drawn from the global node lookup files.
	nodeServices=new ServiceDatabase(this);
    }

 
    // --- Node configuration constants. Accessed from all over the node package.
    // Change them in the config file if needed.

    // ++ External positioning of the node++ :


    public final String NODE_ID() {
	return lookupConfigValue("NODE_ID");
    }

    public final String NODE_LOOKUP_TABLE() {
	return lookupConfigValue("NODE_LOOKUP_TABLE");
    }
    public final String NEIGHBOR_LOOKUP_TABLE() {
	return lookupConfigValue("NEIGHBOR_LOOKUP_TABLE");
    }

    public final String NAME() {
	//return localHost();} // The senderID of the node (for the ProtocolObject). Not what is stored in configObject-> should be consistent.
	return aboutMe.dnsName;
    }

    public final int SERVER_PORT() { // The port to contact the node.
	return aboutMe.port;
    }
    public final String ORGANISATION() { // The provider's id.
	return aboutMe.organisation;
    }
    public final String PASSPHRASE() {
	return lookupConfigValue("PASSPHRASE");
    }
    public final String PGP_TMP_PATH() {
	String pth= lookupConfigValue("PGP_TMP_PATH");
	return ptParser.toPath(pth);
    }
    public final String SCRIPT_PATH(){
	String pth= lookupConfigValue("SCRIPT_PATH");
	return ptParser.toPath(pth);
    }
    public final String TEST_FILE_NAME(){
	// If empty, do live tests. // ="logDump.txt";//="ahEsp3000x1024.tcpdump.output"; // ="ahEspShort.txt";//="test.txt";//  ="ahEspShort.txt"
	return lookupConfigValue("TEST_FILE_NAME");
    }
 
    public final boolean T_IS_VIRTUAL(){
	// Is the t_component test done local or is rsh used.
	String str= lookupConfigValue("T_IS_VIRTUAL");
	return ptParser.toBoolean(str);
    }
    public final String T_NAME(){
	// The dns name of the t_component. //  ="wind.cnds.unibe.ch"; //fermat 
	String str= lookupConfigValue("T_NAME");
	return ptParser.toDNS(str);
    } 

    public final int T_DUMMY_INTERVAL(){
	// In case of local-live (not supported) we use a dummy that generates
	// a packet every T_DUMMY_INTERVAL ms.
	String str= lookupConfigValue("T_DUMMY_INTERVAL");
	return ptParser.toInt(str);
    }


    // ++ Agent environment configuration++ :

    public final int MAX_AGENTS() {
	String str= lookupConfigValue("MAX_AGENTS");
	return ptParser.toInt(str);
    }
    public final int T_SERVER_PORT(){
	String str= lookupConfigValue("T_SERVER_PORT");
	return ptParser.toPort(str);
    }

    // Agent environment:
    public final int Q_LEN(){
	// The len of the queue that bufferes packets for an agent.
	String str= lookupConfigValue("Q_LEN");
	return ptParser.toInt(str);
    }

    public final int CRITICAL_Q_LEN(){	
	// If the queue is full up to here start calling 'emergencyHandle()'
	String str= lookupConfigValue("CRITICAL_Q_LEN");
	return ptParser.toInt(str);
    } 

    // Policy and ressource control info:
    public final boolean ENABLE_RESSOURCE_CONTROL() {
	String str= lookupConfigValue("ENABLE_RESSOURCE_CONTROL");
	return ptParser.toBoolean(str);
    }

    public final long MAX_MEM(){
	// The max size of an agent in bytes.
	String str= lookupConfigValue("MAX_MEM");
	return ptParser.toLong(str);
    }
    public final long MAX_EXEC_TIME(){
	// In millisecs
	String str= lookupConfigValue("MAX_EXEC_TIME");
	return ptParser.toLong(str);
    } 
    public final int TIME_TO_LIVE(){
	 // for agents (in seconds).
	String str= lookupConfigValue("TIME_TO_LIVE");
	return ptParser.toInt(str);
    }

    public final long CONTROL_INTERVAL(){
	// Millisecs to wait between checks.
	String str= lookupConfigValue("CONTROL_INTERVAL");
	return ptParser.toLong(str);
    }
 
    public final double INTOLERABLE_LOSS(){
	// Percentage of dropped packets. //could be done by agent.
	String str= lookupConfigValue("INTOLERABLE_LOSS");
	return ptParser.toDouble(str);
    }
    public final long MAX_BURST_LEN() {
	// Max time between starting the treatment of a packet and stopping.
	String str= lookupConfigValue("MAX_BURST_LEN");
	return ptParser.toLong(str);
    } 
    public final long MIN_SIGNIFICANT_EXEC_LEN() {
	// Only after an agent has executed at least that long, it can be scheduled for killing.
	String str= lookupConfigValue("MIN_SIGNIFICANT_EXEC_LEN");
	return ptParser.toLong(str);
    }
    public final int MAX_TIME_2_NORMAL() {
	// Only after that many intervalls after a killing of an agent, the next agent unfair agent can be killed. 
	String str= lookupConfigValue("MAX_TIME_2_NORMAL");
	return ptParser.toInt(str);
    } 

    // ++ Registry configuration++ :

    public final String CONFIG_PATH() {
	String pth= lookupConfigValue("CONFIG_PATH");
	return ptParser.toPath(pth);
    }
    public final String USER_DB_FILENAME() {
	//Name of file that stores the user profiles.
	return lookupConfigValue("USER_DB_FILENAME");
    }
    // The extension of the filename is dependend of the format (see UserDBManager).
    public final String USER_DB_OBJ_EXT(){
	return lookupConfigValue("USER_DB_OBJ_EXT");
    } 
    public final String USER_DB_TXT_EXT(){
	return lookupConfigValue("USER_DB_TXT_EXT");
    } 
    public final boolean USER_DB_AS_OBJECT(){
	//Stores the user profiles as serialzed
	// objects rather than a human readable version.
	String pth= lookupConfigValue("USER_DB_AS_OBJECT");
	return ptParser.toBoolean(pth);
    }

    // Node services:
    public final long MAX_USES_PER_SERVICE() {
	// For unrestricted node services, this is the max number of uses.
	String str= lookupConfigValue("MAX_USES_PER_SERVICE");
	return ptParser.toLong(str);
    } 



    // --- PUBLIC Services to other node components

    public Registry registry() {
	return registry;
    }

    public ServiceDatabase nodeServices() {
	return nodeServices;
    }

    // location/topology information interface.

    public RouteRecord nextNodeTowards(int ip) {
	// Returns the ID of the next node on the path to the destination ip
	return  routingTable.routingFor(ip);
    }

    RoutingTable routing() {
	// Only for testing purpose.
	return routingTable;
    }

    public NodeRecord aboutMe() {
	return aboutMe;
    }
    public NodeRecord aboutNode(String nodeID) {
	// General (dns-like) lookup
	return nodeLookup.recordOf(nodeID);
    }

    public Vector aboutNodes(Vector nodeIDs) {
	// Convenient list of NodeRecords.
	return nodeLookup.recordsOf(nodeIDs);
    }

    NeighborRecord neighbors() {
	// Returns a list of node IDs of the adjacent nodes or direct users.
	return neighborLookup.allNeighborsOf(NODE_ID());
    }

    NeighborRecord directUsers() {
	NeighborRecord all=neighbors();
	return nodeLookup.directUsers(all);
    }
 
    public NeighborRecord neighborNodes() {
	NeighborRecord all=neighbors();
	return nodeLookup.neighborNodes(all);
    }
    
    public Timer getTimer() {
	return timer;
    }

    // --- Setting the node state ----

    protected void initNodeInfo() {
	// The location/topology info comes from other files delcared in the configfile.
	neighborLookup=new NeighborLookupTable(NEIGHBOR_LOOKUP_TABLE()); // init
	nodeLookup=new NodeLookupTable(NODE_LOOKUP_TABLE()); 
	// the lookup table.
	aboutMe=nodeLookup.recordOf(NODE_ID()); // get my node info.
	if(aboutMe==null) {
	    System.err.println("Didn't find global node file: "+NODE_LOOKUP_TABLE());
	    System.err.print("as declared in my config file!");
	    System.exit(2);
	}
// 	if(!NAME().equals(aboutMe.dnsName)) 
// 	    System.err.println("Warning: inconsistency in DNS configuration! "+NAME() ); // Maybe add a test here: localhost isIn dnsName ...
	if(aboutMe.isUser)
	    System.err.println("Warning: node is declared as user!");
	routingTable=new RoutingTable(this , neighborLookup, nodeLookup );
    }

    protected void initRegistry() {
	registry=new Registry(this);
	registry.configRegistry();
    }

    protected void startNode() {
	System.setSecurityManager(new AgentSecurityManager()); // Starting security.
	initRegistry();

	Server s=new Server(SERVER_PORT(), new UserConnectionHandler(this));
	s.serverLoop();
    }

    public void stopNode() {
	System.out.println("STARTING SHUTDOWN OF THE NODE: deregistering agents:");
	registry.setCongestion(true); // We use this to lock up the node.
	registry.killAllAgents(); // Release the agents in a relative soft way.
	try{Thread.sleep(300);} catch(InterruptedException e) {}; // Give some time to the threads.
	System.out.println("SHUTDOWN NOW.");
	System.exit(0); // Stop the hard way.
    }

    public String localHost() {
	// Looks up the dns name of the local machine. Needed?
	try {
	    InetAddress a=InetAddress.getLocalHost();
	    return a.getHostName();
	} catch (UnknownHostException e) {
	    System.err.println("Node cannot lookup its dns name!");
	    return "UnconnectedNode";
	}
    }

    protected String lookupConfigValue(String key) {
	// Looks into the config object if the configuration parameter
	// to an attibute (key) is there.
	Object value;
	value=configurations.get(key);
	if(value==null) {
	    System.err.println("Node error: missing value for configuration key: "+key);
	    System.exit(7);
	}
	return (String) value; 
    }


    public static void main(String[] args){
	// Main entrance: starts one complete node.
	// usage: node.Node [confFile]
	Node node;
	if (args.length>0) { 
	    System.out.println("Starting node with arguments: "+args[0]);
	    node=new Node(args[0]);

	} else {
	    System.out.println("Starting node without arguments!");
	    node=new Node(defaultConfigName);
	}

	node.startNode();
    }
}
