package node;

import java.io.*;
import java.net.*;
import utils.ObjectQueue;

import capsule.Agent;
import clientserver.Result;
import filter.*;

public class Receiver extends EnvironmentComponent {
    // Receives the T-component output and creates IPPacket Objects for
    // one Agent.
    protected ObjectQueue queue;

    protected Socket socket;
    protected int port;
    protected PacketFactory factory=new PacketFactory();
    protected int nrOfPckts;
    protected Filter filter;

    protected boolean stopped=false; // Synchronisation. If the agent stops at 'init', then
    // the receiver is probably not started yet, but allready stopped!

    protected DataInputStream input=null; //Received data can be read from here.

    protected long connectionStart; // for performance tests.

    public Receiver() {}

    public Receiver (int p, ObjectQueue q, Filter f, Node nod, Agent a) {
	if(portOK(p)) port=p; else {
	    System.err.println("T_C: Illegal port, using 3330 instead.");
	    port=3330;
	}
	init(nod,a);
	queue=q;
	filter=f;
	nrOfPckts=f.packets().getNr();
    }

    void setSocket(Socket socket) {
	// Socket already opened by external party (T_Configurator);
	this.socket=socket;
	setInput(socket);
    }


    public void run() {
	Thread.currentThread().setPriority(Thread.MAX_PRIORITY-1);
	// A high priority guarantees, that a DoS attack cannot
	// slow down the packet reception which is used as a measurement
	// of the cpu consumption of the agents (see RessourceController).
	if(!initSocket()) return;
	connectionStart=System.currentTimeMillis(); // performance tests.
	receive();
	long connectionEnd=System.currentTimeMillis();
	System.out.println("T_C: Connection ended after: "+(connectionEnd-connectionStart)+" ms.");
    }
    synchronized void stop(boolean informRegistry) {
	// for external stopping the packetdelivery.
	// maybe the thread could be stopped gracefully?

	//stopped=true;
	closeSocket();
	super.stop(informRegistry);
    }

    protected  boolean initSocket() {
	if (stopped) return false;
	if(input != null) return true; // external socket is used, no need to open the server.
	boolean success=true;
	try {
	    ServerSocket ss=new ServerSocket(port);
	    System.out.println("T_C: Listening for connection on port: "+ss.getLocalPort());
	    socket=ss.accept();
	    System.out.println("T_C: Accepting connection.");
	    setInput(socket);
	} catch(IOException e) {
	    System.err.println("T_C: Error in initSocket(): "+e);
	    success=false;
	    node.registry().stopAgent(agent, new Result("Can't reach T_Component...try again!"));
	}
	return success;
    }

    protected void setInput(Socket socket) {
	try {
	    input=new DataInputStream(socket.getInputStream());
	} catch(IOException e) {
	    System.err.println("Receiver: could not set socket: "+socket+" because: "+e);
	}
    }

    protected synchronized void closeSocket() {
	try {
	    if(socket!=null) // Socket may be null, if the Receiver is stopped before it really started.
		socket.close();
	    System.out.println("T_C: Server socket closed.");
	} catch(IOException e) {
	    System.err.println("T_C: Error in closeSocket(): "+e);
	}
    }

    protected void receive() {
	if(socket == null) return;
	for (int i=0; (nrOfPckts<0) || (i<nrOfPckts); i++) {
	    // Note that a negative i means go on forever.	   
	    int len=readDataLen();
	    //System.out.println("Receiving a header with "+len+" data bytes.");
 	    if(len<0) { // if an error occured
 		closeSocket();
 		break;
 	    }
	    else {
		//System.out.println(factory.create(input,len));
		IPPacket ip=factory.create(input,len);
		queue.put(ip);
	    }
	}
    }

    protected int readDataLen() {
	// The first 2 bytes encode the length of ip data transmitted.
// 	int len=0;
// 	try {
// 	    len=input.read();
// 	    len<<=8;
// 	    len+=input.read();
// 	} catch(IOException e) {
// 	    System.err.println("Error in readDataLen(): "+e);
// 	    len=-1;
// 	}
// 	return len; // the ip bytes transmitted.

	int len=0, h;
	try {
	    h=input.read(); //quickndirty
	    if(h<0) h+=256;
	    h<<=8;
	    len=h;
	    h=input.read();
	    if(h<0) h+=256;
	    len+=h;
	} catch(IOException e) {
	    System.err.println("T_C: Error in readDataLen(): "+e);
	    len=-1;
	}
	return len; // the ip bytes transmitted.
    }

    protected static boolean portOK(int p) {
	return ((p>1023) && (p<65536));
    }


    public static void main(String[] args) {
// 	int runLength=5998; // packets in ahEsp trace: 5998
// 	ObjectQueue q=new ObjectQueue();
// 	Meter meter;
// 	//meter=new ByteFrequencyMeter(10,true); // ~225 kbps, significant ok (73 fails auf 4170 tests)
// 	//meter=new FastByteFrequencyMeter(10,false); //~270 kbps
// 	//meter=new DebugMeter(runLength);
// 	meter=new FastestMeter(1); // 38558 ms , ca 170 KBps (die bps sachen sind jetzt falsch.
// 	//meter=new IPSecConsistencyMeter(1); // the lens and proto fields are ok!
// 	//meter=new FastRunLengthMeter(1,false); //~270 kbps
// 	//meter=new BandwidthMeter(10);
// 	//meter=new IPConsistencyMeter(1); // all kinds of checks e.g. header cksum  36714 ms

// 	Agent a=new Agent("Housi",  meter,  q,  runLength);
// 	Receiver r=new Receiver(3330, q);
// 	(new Thread(r)).start();
// 	(new Thread(a)).start();

    }
}
