package node.services;

import node.*;
import clientserver.*;
import filter.*;
import utils.*;
import topology.NodeRecord;

import java.util.*;
import java.net.*;

public class ChangeChannel extends DeliveringServiceHandler {
    // A client can open a new connection over which HomeChannel will send then.
  
    protected ExecutionEnvironment execEnv;
    protected ServerPOSenderReceiver posr;
    protected HomeChannel homeChannel=null;

    protected String receiverID; // Temp data: (set in parseArgumentsToIVs())
    protected String server; 
    protected int port;
    protected boolean encrypt;
    protected boolean auth;
    protected byte encoding;

    ServiceHandler newServiceHandlerFor(ExecutionEnvironment ev, Node n) {
	return new ChangeChannel(n, ev, type);
    }

    ChangeChannel(int type) {
	// Non-complete constructor , which is then used as a factory.
	this.type=type;
    };
 
    ChangeChannel(Node n, ExecutionEnvironment ev, int typ) {
	// Only the ServiceDatabase uses the constructor, to assure the type consistency.
	initDelivery(n,ev,typ);
	execEnv=ev;
    }


    boolean startService(Service req) {
	Service reply=new Service(req);
	Vector v=new Vector(1);

	if(!parseArgumentsToIVs(req) || !policyCheck())
	    return false; // Parsing errors lead to the termination of the agent.
	if(!openChannel()){
	    returnChannelChangeState(reply, false);  // Signal to the agent that the request has failed.
	    return true; // Don't kill the agent for connection problems.
	}
	if(!changeHomeChannel()) // Install the new home channel.
	    return false;
	returnChannelChangeState(reply, true);  
	return true;
    }


    void stopService() {
	    // HomeChannel should do the closing.
    }

    protected boolean openChannel() {
	// Opens the channel by re-init the posr.
	Socket socket;
	try {
	    socket=new Socket(server, port);
	} catch (Exception e) {System.err.println(e); socket=null; return false;}
	posr=new ServerPOSenderReceiver(socket, node);
	posr.setEncoding(encoding, receiverID); // Here we set the encryption/auth.
	// Receiver is the owner of the agent.
	return true;
    }

    protected boolean parseArgumentsToIVs(Service req) {
	// Parsing sets the temp IVs.
	Vector v=req.arguments();
	Object arg;
	if(v==null) return false;
	Enumeration e=v.elements();
	if(v.size()<4) {
	    System.err.println("ChangeChannel: not enough arguments for the change channel service: "+v);
	    return false;
	}
	arg=e.nextElement();  // DNS
	if(!(arg instanceof String)) {
	    setDefaultReceiverInfo();
	} else
	    server=(String) arg;
	arg=e.nextElement(); // Port
	if(!(arg instanceof Integer)) {
	    System.err.println("ChangeChannel: expecting Port Integer instead of: "+arg);
	    return false;
	}
	port=((Integer) arg).intValue();
	arg=e.nextElement(); // Encryption
	if(!(arg instanceof Boolean)) {
	    System.err.println("ChangeChannel: expecting Enc. Boolean instead of: "+arg);
	    return false;
	}
	encrypt=((Boolean) arg).booleanValue();
	arg=e.nextElement(); // Authentication
	if(!(arg instanceof Boolean)) {
	    System.err.println("ChangeChannel: expecting Authentication Boolean instead of: "+arg);
	    return false;
	}
	encrypt=((Boolean) arg).booleanValue();
	setEncoding(); // Calc the encoding from encrypt and auth
	if(encrypt) { // ReceiverID only relevant if encrypting.
	    arg=e.nextElement();  // ReceiverID
	    if(!(arg instanceof String)) {
		setDefaultReceiverInfo(); // overrides also the 'server'...
	    }
	    else receiverID=(String) arg;
	}
	return true;
    }

    protected void setDefaultReceiverInfo() {
	// If the agent didn't provide receiver info, we assume that there's an entry
	// for the owner of the agent in the node lookup table.
	receiverID=execEnv.owner();
	NodeRecord nr=node.aboutNode(receiverID);
	if(!nr.isUser) System.err.println("Warning: change channel to a node instead of a user.");
	server=nr.dnsName;
	
    }

    protected void setEncoding() {
	// Translates the high-level crypto spec (booleans) into a encoding id.
	if(encrypt) encoding=ProtoConsts.FULL_PGP_v263i; 
	else
	    if(auth) encoding=ProtoConsts.SIGN_PGP_v263i; 
	    else encoding=ProtoConsts.PLAIN;
	// Note: enc && !auth is not supported.
    }
    

    protected boolean policyCheck() {
	// Checks if dns exists and if corresponding ip matches the filters.
	InetAddress ip;
	int ipi;
	try {
	    ip=InetAddress.getByName(server);
	} catch(UnknownHostException e) {
	    System.err.println("Unknown home address: "+server);
	    return false;
	}
	ipi=AddressCalculator.parseIPaddress(ip.getHostAddress());
	Filter res, filter=new Filter(new ListOfRanges(ipi), new ListOfRanges(ipi),new ListOfRanges(port), new All() , -1, -1); // This filter describes the requested connection.
	res=execEnv.policy().getFirstMatchingFilter(filter); // Check if the request matches
	// the policy. Basically, the connection is allowed iff the agent is
	// allowed to listen to traffic coming and going to this port of this server.
	if (res==null) { // Policy does not allow to connect there.
	    System.err.println("Policy forbids connection to: "+server+":"+port);
	    return false;
	}
	return true;
    }

    protected boolean changeHomeChannel() {
	if (serviceList==null) {
	    System.err.println("Error in ChangeChannel: the list iv was not initialized");
	    return false;
	}
	if (homeChannel == null) { // Lazy init
	    ServiceListEntry entry=serviceList.getServiceListEntry(ServiceConsts.SEND());
	    homeChannel= (HomeChannel) entry.handler();
	}
	homeChannel.replacePOSR(posr); // Install the new channel.
	posr.send(createCallBackProtoObj());
	// Init the protocol by signaling who wants the connection (senderID->agentID).
	// Maybe this should be done via homeChannel?
	return true;
    }

    protected ProtocolObject createCallBackProtoObj() {
	return new ProtocolObject(node.NODE_ID(), ProtoConsts.PLAIN ,  ProtoConsts.CALLBACK, createCallBack());
    }

    protected CallBack createCallBack() {
	return new CallBack(execEnv.agentID(), execEnv.hopCount(), node.aboutMe());
    }

    protected void returnChannelChangeState(Service reply, boolean ok) {
	// Generate a return value in response to the service call.
	Vector v=new Vector(1);
	v.addElement(new Boolean(ok)); 
	reply.setArgs(v);
	deliveryQueue.put(reply);
    }

}
