package topology;

import node.Node;
import clientserver.*; // For sending routing updates

import utils.*;

import java.util.*;

public class RoutingTable extends TopologyTable {
    // Contains all infos about location and name of all nodes.

    protected String nodeID; // the ID.
    protected NeighborLookupTable neighborT;
    protected NodeLookupTable nodeT;
    protected Node node;

    protected Vector neighbors; // All NodeRecords of neighbors (for convenience).
    protected Vector users; // All NodeRecords of direct users (for convenience).

    // public RoutingTable peer; //for testing

    public RoutingTable(Node node,NeighborLookupTable nei, NodeLookupTable nol) {
	this.node=node;
	this.nodeID=node.NODE_ID();
	nodeT=nol;
	neighborT=nei;
	initFromTables();
	startRoutingReceiver();
	try {Thread.sleep(8000);}catch(InterruptedException e){};// Make sure all
	// other routing tables are running.
	startUpdate();
	
    }

//     public RoutingTable(String nodeID,NeighborLookupTable nei, NodeLookupTable nol) {
// 	// Obsolete
// 	node=nodeID;
// 	nodeT=nol;
// 	neighborT=nei;
// 	initFromTables();
//     }


    protected void startRoutingReceiver() {
	NodeRecord me=nodeT.recordOf(nodeID); // Lookup my routing server port.
	Server s=new Server(me.routingPort, new RouteUpdateConnectionHandler(this, node));
	ServerStarter start=new ServerStarter(s);
	(new Thread(start)).start();
    }

    public synchronized void routeUpdate(Vector newRoutes, String senderID) {
	// Incorporates the routing algorithm (shortest hop-count).
	//System.out.println("Updating new routes: "+newRoutes);
	RouteRecord rr, old;
	Vector updated=new Vector(); // For propagating the updates
	Object key;

	for(Enumeration e=newRoutes.elements(); e.hasMoreElements();) {
	    rr=(RouteRecord) e.nextElement();
	    key=rr.key;
	    old=(RouteRecord) table.get(key);

	    if (old==null) {
		table.put(key, rr);
		updated.addElement(rr);
	    } else
		if(old.hopCount>rr.hopCount) {
		    table.remove(key);
		    table.put(key, rr);
		    updated.addElement(rr);
		}
	}
	if(updated.size()>0)
	    informNeighbors(updated, senderID);
    }

    protected void informNeighbors(Vector updated, String senderID) {
	//System.out.println("Inform Neighbors: "+updated);

	RouteRecord u;
	Vector forNeighbor=new Vector();
	for(Enumeration e=updated.elements(); e.hasMoreElements();) {
	    // Update the routing fields for the next hop.
	    u=(RouteRecord) e.nextElement();
	    forNeighbor.addElement(broadCastRecord(u));
	}
	if(neighbors!=null) { // In the init phase this may happen..
	    NodeRecord receiver;
	    for(Enumeration e=neighbors.elements(); e.hasMoreElements();) {
		// To all neighbor nodes:
		receiver= (NodeRecord)  e.nextElement();
		if(!senderID.equals(receiver.key)) // Don't send it back to where it came from.
		    sendRoutingUpdate(receiver , forNeighbor);
	    }
	}
    }

    protected void sendRoutingUpdate(NodeRecord to, Vector updates) {
	RoutingClient rc=new RoutingClient(to);
	rc.send(new RoutingUpdate(updates), node.PASSPHRASE());
    }


    protected RouteRecord broadCastRecord(RouteRecord rr) {
	// Constructs a new adapted record for sending to the peering nodes.
	RouteRecord ret=new RouteRecord(rr.ip, nodeID); // The route goes throu nodeID.
	ret.hopCount=rr.hopCount+1; // It's one hope more.
	return ret;
    }

    public RouteRecord routingFor(int ip) {
	String key=RouteRecord.calcKey(ip);
	return (RouteRecord) table.get(key);
    }

//     protected Object calcKey(RouteRecord rr) {
// 	// Hashtable needs an Object as the key. We use the 10.11.13 form.
// 	String key=calcKey(rr.ip);
// 	rr.key=key;
// 	return key;
//     }
//     protected String calcKey(int key) {
// 	// Hashtable needs an Object as the key. We use the 10.11.13 form.
// 	return AddressCalculator.toString(key);
//     }

    protected synchronized void initFromTables() {
	NeighborRecord nr=neighborT.allNeighborsOf(nodeID);
	neighbors=allIn(nr,false); // Vector of NodeRecords of direct users.
	users=allIn(nr,true);// Vector of NodeRecords of all neighbor nodes.
    }

    protected synchronized void startUpdate() {
	Vector newRoutes=getRoutesOfDirectUsers(users);
	routeUpdate(newRoutes, ""); // Uses the standard way to set the routes. Use empty sender
    }

    protected Vector allIn(NeighborRecord nr, boolean users) {
	// Creates a separate list with either all users or all nodes.
	NeighborRecord only=nodeT.neighbors(nr,users);
	if(only==null) return new Vector();
	return nodeT.recordsOf(only.neighbors);
    }

    protected Vector getRoutesOfDirectUsers(Vector userRecords) {
	Vector newRoutes=new Vector();
	NodeRecord userR;
	for(Enumeration e=userRecords.elements(); e.hasMoreElements();) {
	    userR=(NodeRecord) e.nextElement();
	    newRoutes.addElement(new RouteRecord(userR.ip, nodeID));
	    // Here we set the route to us, since we are the last hop towards
	    // this (directly connected) user.
	}
	return newRoutes;
    }


    public  String toString() {
	return "RoutingTable: "+super.toString();
    }

    public static void main(String[] args) {
	//	RoutingTable genf=new RoutingTable("<admin@cui.unige.ch>",new NeighborLookupTable("./neighbor.global.config"), new NodeLookupTable("./nodes.global.config"));
// 	RoutingTable bern=new RoutingTable("<admin@iam.unibe.ch>",new NeighborLookupTable("./neighbor.global.config"), new NodeLookupTable("./nodes.global.config"));

// 	genf.peer=bern;
// 	bern.peer=genf;

// 	genf.initFromTables();
// 	bern.initFromTables();

// 	System.out.println(genf.routingFor(AddressCalculator.parseIPaddress("130.92.64.15")));

    }
    
}
