package utils;

import java.io.*;

public class OneTimeInputStream extends InputStream {
    // Gives access to a temorary file created by a pgp script.
    // This class delegates most functionality to a FileInputStream.
    // However, once the stream is closed, all files matching
    // 'filename' are removed.
    protected String filename;
    protected String extension;
    protected String path= "."; // Add another path here.
    protected InputStream fin=null;
    protected boolean failed=false;
    
    public OneTimeInputStream(String pth, String fn, String ext) {
	path=pth;
	filename=fn;
	extension=ext;
	try {
	    fin=new FileInputStream(path+fn+ext);
	} catch (IOException e) {
	    System.err.println(e);
	    fin=new ByteArrayInputStream(new byte[0]); // Empty stream.
	    failed=true;
	}
    }

    public String getFilename() {
	return filename;
    }

    // The delegated methods:
    public int available() throws IOException {
	return fin.available();
    }
    public synchronized void mark(int i) {
	fin.mark(i);
    }
    public boolean markSupported()  {
	return fin.markSupported();
    }
    public int read() throws IOException {
       return fin.read();
    }
    public int read(byte[] b) throws IOException {
       return fin.read(b);
    }
    public int read(byte[] b, int off, int len) throws IOException {
       return fin.read(b,off,len);
    }
    public synchronized void reset() throws IOException {
	fin.reset();
    }
    public long skip(long n) throws IOException {
	return fin.skip(n);
    }
    
    // Closing the stream removes the temp files.
    public void close() throws IOException {
	fin.close();
	File dir=new File(path);
	removeFiles(dir.list());
    }

    protected void removeFiles(String[] list) {
	File f=null;
	String name;
	int fnSize=filename.length();
	for (int i=0, j=0; i< list.length; i++) {
	    name=list[i];
	    if (filename.regionMatches(0,name,0, fnSize)) { 
		// Remove only files matching the filename.
		f=new File(path, list[i]);
		if (!f.delete()) System.err.println("Deletion of pgp temporary file failed!");
	    }
	}
    }

    public static void main(String[] args) {
	// Testing
	OneTimeInputStream s= new OneTimeInputStream("./","0",".pgp");
	try{
	    while(s.available()>0) {
		System.out.print(s.read()+" ");
	    }
	    s.close();
	} catch (IOException e) {System.err.println(e);}
    }

}
					   
