/* Written by Manuel Gnter. All rights reserved, no warranties */

package utils;

import java.io.*;
import java.util.*;

public class PGPEncoder263i {
    // This class can encode and decode and authenticate messages using
    // pgp. The interaction with pgp is script based.

    // Call sequence: sign & encrypt:   encode() or:
    //                sign only:        sign()
    //                decrypt/decode:   decode()
    //                verify sign:      authenticate()
    //                who has signed:   getSigningParty
    //                was it encrypted: wasEncrypted() 
    // This is the adaption for pgp v.2.6.3i
    // The batch mode works (does not block).
    // Decryption of a corrupted file is not done which leads to a FileIO Exception
    // that is caught -> no problem.
    // Note: some code duplication could be removed, if time available.

    protected static final String decScriptName="decodePGP263i";
    protected static final String sigScriptName="signPGP263i";
    protected static final String encScriptName="encodePGP263i";
    protected static final String crypt_extension=".pgp"; // Should be compatible with the script.
    protected static final String dec_extension=".clr"; // Should be compatible with the script.
    protected static final String auth_extension=".aut";
    protected static final String scriptPath="/home/mguenter/rvs/Agents/utils/Scripts/";

    protected String tmpPath; // The path where the temporary enc/dec files are stored.
    protected static int id=0; // Changed to static!
    protected String encScriptStart; // The scripts that en-/decrypt.
    protected String sigScriptStart;
    protected String decScriptStart;

    protected String clientName; // For unique id of the file if serveral processes are running.

    protected boolean wasEncrypted=false; 

    // The public methods: ----------------------

    public PGPEncoder263i(String tmpP) {
	tmpPath=tmpP;
    }

    public PGPEncoder263i(String tmpP, String client) {
	clientName=client;
	tmpPath=tmpP;
    }

    public void setEncoder(String client) {
	clientName=client;
    }

    public OneTimeInputStream encode(InputStream s, String sender, String receiver, String PPhSenderKey) {
	// Encryption and signing
	return prepareAndStartScript(s, encScriptName , receiver, sender, PPhSenderKey, "",  crypt_extension);
    }


    public OneTimeInputStream sign(InputStream s, String sender, String PPhSenderKey) {
	// Signing only.

	return prepareAndStartScript(s, sigScriptName , "", sender, PPhSenderKey, "",  crypt_extension);
    }


    public OneTimeInputStream decode(InputStream s, String PPhSenderKey) {
	// Decryption/decoding (for signed only).
	return prepareAndStartScript(s, decScriptName , "", "", PPhSenderKey, crypt_extension,  dec_extension);
    }



    public boolean authenticate(OneTimeInputStream decoded) {
	// After decoding, call this method on the decoded stream to
	// check the signature. As a side-effect, it sets 'wasEncrypted' so
	// that the client can differ between signed-only and encrypted.
	// We analyse the pgp output file which the decoding process has stored
	// in filename.auth.
	if (decoded==null) return false;
	int firstLetter=0;
	int goodFirst=71; // PGP output should start with 'G' for "Good signature"
	try {
	    FileInputStream fin=openFileOf(decoded);
	    skipLines(fin,3); 
	    firstLetter=fin.read();
	    if (firstLetter==goodFirst) { // Sign-only output looks different.
		wasEncrypted=false;
		fin.close();
		return true;
	    }
	    skipLines(fin,4);
	    firstLetter=fin.read();
	    if (firstLetter==goodFirst) { // Signed & encrypted.
		wasEncrypted=true;
		fin.close();
		return true;
	    }
	    fin.close();
	} catch (IOException e) {
	    System.err.println(e);
	}
	return false;
    }

    public String getSigningParty(OneTimeInputStream decodedNauthed)  {
	// This method extracts the email address as ID from the pgp
	// authentication output, thus it works similar as the 'authenticate' method.
	// It looks for a statement like this:  
	// "Broker Bern <bernBroker@iam.unibe.ch>"
	// Works only if 'authenticate()' has already been called!!!

	if (decodedNauthed==null) return null;
	String line="";
	FileInputStream fin=openFileOf(decodedNauthed);
	if (fin != null) try {
	    if (wasEncrypted)
		skipLines(fin,7); // The output format is other than in v5.
	    else
		skipLines(fin,3);
	    line= readLine(fin);
	    fin.close();
	} catch (IOException e) {
	    System.err.println(e);
	}
	String ret="";
	StringTokenizer parser= new StringTokenizer(line,"\""); // This depends on the .aut output
	try{
	    parser.nextToken(); 
	    ret=parser.nextToken();
	} catch (NoSuchElementException e) {System.err.println(e);}
	return ret;
    }

    public boolean wasEncrypted() {
	// After decode() and authenticate() has been called successfully,
	// this can be used to differ between sign&encrypt and signed-only.
	return wasEncrypted;
    }

    // File IO Helpers ------------------

    protected FileInputStream openFileOf(OneTimeInputStream oin) {
	// This is specifically used for authentication (find the pgp auth. output file).
	String filename=oin.getFilename();
	FileInputStream fin=null;
	try {
	    fin=new FileInputStream(tmpPath+filename+auth_extension);
	} catch (IOException e) {
	    System.err.println(e);
	}
	return fin;
    }

    protected String readLine(FileInputStream fin)  throws IOException {
	StringBuffer buf= new StringBuffer(64);
	int tmp=0;
	while(fin.available()>0) {
	    tmp=fin.read();
	    buf.append(new Character((char) tmp)); // This works only for ascii.
	    if (tmp==10) break;
	}
	return new String(buf);
    }

    protected void writeToFile(String filename, InputStream data) {
	FileOutputStream fout=null;
	try {
	    fout=new FileOutputStream(filename);
	} catch (IOException e) {System.err.println(e); System.exit(-1);}

	try {
	    while (data.available()>0) {
		fout.write(data.read());
	    }
	    data.close(); // If data was a OneTimeInputStream this is nice.
	    fout.close();
	} catch (IOException e) {System.err.println(e);}
    }



    // Parsing Helpers ------------------



    protected void skipLines(FileInputStream fin, int lines) throws IOException {
	int tmp=0;
	while((fin.available()>0) && (lines>0)) {
	    tmp=fin.read();
	    if (tmp==10) lines--; // The 'newline's.
	}
    }


    // Unique filenames ------------------

    protected static synchronized int getNewID() {
	return id++; // Should be reset to 0 in case of overflow.
    }


    protected String createUniqueFileName() {
	int nameSize=8; // This limits us to 100mio encryptions. Not nice.
	StringBuffer buf= new StringBuffer(nameSize);
	String name=String.valueOf(getNewID());
	int namlen=clientName.length();
	if (namlen>10) namlen=10;
	int namstrt=(namlen-4);
	if (namstrt<0) namstrt=0;
	
	name+=clientName.substring(namstrt,namlen); // use some letters to create a unique name.
	int nrOfLeadingZeroes=nameSize-name.length();
	for (int i=0; i<nameSize; i++) buf.append('0');
	buf.append(name);
	return new String(buf);
    }

    // Skript helpers ----------------------

    protected OneTimeInputStream prepareAndStartScript(InputStream s, String scriptName, String id1, String id2, String PPhSenderKey, String inExtension,  String outExtension) {
	// Composes the script and its arguments and runs it.

	String filename=createUniqueFileName();
	String fullname=tmpPath+filename;
	writeToFile(fullname+inExtension,s); // Temporary store the data.
	
	// Start pgp script on the file
	StringBuffer script=new StringBuffer(128);
	preparePreamble(script);

	script.append(scriptName).append(' ').append(fullname).append(' ').append(id1).append(' ').append(id2).append(' ').append(PPhSenderKey);
      	if(executeScript(new String(script))) {
	    //try{System.in.read();}catch(IOException ie){} // Wait (testing manipulations of the files)
	} else { 
	    System.err.println("PGP script failure for: "+scriptName);
	    try{s.close();} catch(IOException e){};// debugg 29.11.2000
	    return null; // Debugg 13.11.2000
	}
	return new OneTimeInputStream(tmpPath,filename,outExtension); // Return something anyways (can be null!).
    }

    protected void preparePreamble(StringBuffer buf) {
	buf.append("/bin/sh ").append(scriptPath);
    }

    protected boolean executeScript(String script) {
      	return SkriptExec.executeScript(script, true)==0; // true if script was successfull.
    }


    // Main (testing) -------------------------

    public static void main(String[] args) {
	PGPEncoder263i e= new PGPEncoder263i("./","PREABMLE");
	String plaintext="Hey di how di doodle haaaww";
	ByteArrayInputStream in= new ByteArrayInputStream(plaintext.getBytes());
	
	//OneTimeInputStream s= e.encode(in,"admin@iam.unibe.ch","admin@cui.unige.ch","iamPPh:iam.unibe.ch"); 
	OneTimeInputStream s= e.sign(in,"admin@iam.unibe.ch","iamPPh:iam.unibe.ch"); 

	OneTimeInputStream s2= e.decode(s,"cuiPPh:cui.unige.ch");// xxx change keys back!

	System.err.println("Authentication: "+e.authenticate(s2));
	System.err.println("Signing Party: "+e.getSigningParty(s2));


	try{
	    while(s2.available()>0) {
			System.out.print((char)s2.read());
	    }
		System.out.println();
	    s2.close();
	} catch (IOException ioe) {System.err.println(ioe);}
    }

}
